# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import argparse
from typing import Any, List, Tuple

import numpy as np


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="PPD plot",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="This script plots ppd plots.",
    )
    parser.add_argument(
        "--data",
        nargs="+",
        help="data files paths (one or more). Only .hdf5 files are supported.",
        type=str,
        required=True,
    )
    parser.add_argument(
        "--dir",
        type=str,
        required=True,
        help="directory to save the plots",
    )
    parser.add_argument(
        "--prefix",
        help="prefix for the output file",
        type=str,
    )

    pretty_group = parser.add_argument_group("Pretty Options")
    pretty_group.add_argument(
        "--titles",
        nargs="+",
        type=str,
        help="titles for the plots in order of the parameters",
    )
    pretty_group.add_argument(
        "--x-scale",
        help="scale of the x-axis",
        default="linear",
        choices=["linear", "log"],
        type=str,
    )
    pretty_group.add_argument(
        "--y-scale",
        help="scale of the y-axis",
        default="linear",
        choices=["linear", "log"],
        type=str,
    )
    pretty_group.add_argument(
        "--x-range",
        nargs=2,
        action="append",
        type=float,
        help="range of the x axis plot in the form of start end for each parameter",
    )
    pretty_group.add_argument(
        "--y-range",
        nargs=2,
        action="append",
        type=float,
        help="range of the y axis plot in the form of start end for each parameter",
    )
    pretty_group.add_argument(
        "--font-size",
        default=16,
        type=int,
        help="font size for the x,y axis labels and titles",
    )
    pretty_group.add_argument(
        "--x-labels",
        nargs="+",
        type=str,
        help="labels for the x axis in order of the parameters",
    )
    pretty_group.add_argument(
        "--y-labels",
        nargs="+",
        type=str,
        help="labels for the y axis in order of the parameters",
    )
    pretty_group.add_argument(
        "--size",
        help="size of the ppd plot in inches",
        nargs=2,
        default=(12, 12),
        type=float,
    )
    pretty_group.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    pretty_group.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    pretty_group.add_argument(
        "--median-color",
        help="color of the median line",
        type=str,
        default=r"#7B8794",
    )
    pretty_group.add_argument(
        "--ppd-color",
        help="color of the ppd line",
        type=str,
        default=r"#b41f78",
    )
    pretty_group.add_argument(
        "--ninety-ci-color",
        help="color of the 90 percentile CI",
        type=str,
        default=r"#C5C9C7",
    )
    pretty_group.add_argument(
        "--fifty-ci-color",
        help="color of the 50 percentile CI",
        type=str,
        default=r"#BBF90F",
    )
    pretty_group.add_argument(
        "--median-alpha",
        help="alpha of the median line",
        type=float,
        default=0.8,
    )
    pretty_group.add_argument(
        "--ppd-alpha",
        help="alpha of the ppd line",
        type=float,
        default=1.0,
    )
    pretty_group.add_argument(
        "--ninety-ci-alpha",
        help="alpha of the 90 percentile CI",
        type=float,
        default=0.5,
    )
    pretty_group.add_argument(
        "--fifty-ci-alpha",
        help="alpha of the 50 percentile CI",
        type=float,
        default=0.7,
    )
    pretty_group.add_argument(
        "--median-linestyle",
        help="linestyle of the median line",
        type=str,
        default="--",
    )
    pretty_group.add_argument(
        "--grid",
        help="show grid",
        action="store_true",
    )
    pretty_group.add_argument(
        "--grid-which",
        help="which grid to show",
        type=str,
        default="both",
        choices=["both", "major", "minor"],
    )
    pretty_group.add_argument(
        "--grid-alpha",
        help="alpha of the grid",
        type=float,
        default=0.7,
    )
    pretty_group.add_argument(
        "--grid-linestyle",
        help="linestyle of the grid",
        type=str,
        default="--",
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )

    return parser


def get_domain(
    domain_as_read: Any,
) -> List[Tuple[float, float, int]]:
    """Get the domain of the model per axis.

    Parameters
    ----------
    domain_as_read : Any
        The domain of the model per axis.

    Returns
    -------
    List[Tuple[float, float, int]]
        The domain of the model per axis.
    """
    return [(float(s), float(e), int(n)) for s, e, n in domain_as_read]


def get_utf8_decoded_headers(headers_as_read: Any) -> List[str]:
    """Get utf-8 decoded headers.

    Parameters
    ----------
    headers_as_read : Any
        The headers as read from the file.

    Returns
    -------
    List[str]
        The utf-8 decoded headers.
    """
    return [h.decode("utf-8") for h in headers_as_read]


def get_quantiles(data: np.ndarray, quantiles: np.ndarray) -> np.ndarray:
    """Get the quantiles of the data.

    Parameters
    ----------
    data : np.ndarray
        The data to get the quantiles of.
    quantiles : np.ndarray
        The quantiles to get. The quantiles should be between 0 and 1.

    Returns
    -------
    np.ndarray
        The quantiles of the data.
    """
    return np.quantile(data, quantiles, axis=1)


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    import h5py
    from matplotlib import pyplot as plt
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure

    from kokab.utils.ppd import get_all_marginals

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    prefix = "" if args.prefix is None else args.prefix

    x_range = args.x_range
    y_range = args.y_range
    font_size = args.font_size
    x_labels = args.x_labels
    y_labels = args.y_labels
    titles = args.titles

    quantiles = [
        0.05,  # 90% CI
        0.25,  # 50% CI
        0.5,  # median
        0.75,  # 50% CI
        0.95,  # 90% CI
    ]

    axes: List[Tuple[Figure, Axes]] = []

    for j in range(len(args.data)):
        with h5py.File(args.data[j], "r") as f:
            domains = get_domain(f["domains"])
            headers = get_utf8_decoded_headers(f["headers"])
            marginals = [f["marginals"][head][:] for head in headers]
            ppd = f["ppd"][()]

        marginal_ppds = get_all_marginals(ppd, domains)

        if not axes:
            axes = [plt.subplots(figsize=args.size) for _ in range(len(headers))]

        for i in range(len(headers)):
            _, ax = axes[i]
            head = headers[i]
            marginal = marginals[i]
            marginal_ppd = marginal_ppds[i]
            domain = domains[i]

            start, end, num_points = domain
            quant = get_quantiles(marginal, quantiles)
            xx = np.linspace(start, end, num_points)

            ax.plot(
                xx,
                quant[2],
                # label="Median",
                color=args.median_color,
                alpha=args.median_alpha,
                linestyle=args.median_linestyle,
            )
            ax.plot(
                xx,
                marginal_ppd,
                # label="PPD",
                color=args.ppd_color,
                alpha=args.ppd_alpha,
            )
            ax.fill_between(
                xx,
                quant[0],
                quant[-1],
                alpha=args.ninety_ci_alpha,
                # label="90\% CI" if args.use_latex else "90% CI",
                color=args.ninety_ci_color,
            )
            ax.fill_between(
                xx,
                quant[1],
                quant[-2],
                alpha=args.fifty_ci_alpha,
                # label="50\% CI" if args.use_latex else "50% CI",
                color=args.fifty_ci_color,
            )
            ax.set_yscale(args.y_scale)
            ax.set_xscale(args.x_scale)
            if titles is not None:
                if i < len(titles):
                    ax.set_title(titles[i], fontsize=font_size)
            if x_labels is None:
                ax.set_xlabel(head, fontsize=font_size)
            else:
                if i < len(x_labels):
                    ax.set_xlabel(x_labels[i], fontsize=font_size)

            if y_labels is None:
                ax.set_ylabel(f"ppd({head})", fontsize=font_size)
            else:
                if i < len(y_labels):
                    ax.set_ylabel(y_labels[i], fontsize=font_size)

            if x_range is not None and i < len(x_range):
                ax.set_xlim(x_range[i][0], x_range[i][1])
            if y_range is not None and i < len(y_range):
                ax.set_ylim(y_range[i][0], y_range[i][1])

    for i in range(len(axes)):
        fig, ax = axes[i]
        filename = f"{args.dir}/{prefix}{headers[i]}_ppd_plot"
        if len(axes) > 1:
            filename += f"_{i}.pdf"
        else:
            filename += ".pdf"

        if args.grid:
            ax.grid(
                args.grid,
                which=args.grid_which,
                linestyle=args.grid_linestyle,
                alpha=args.grid_alpha,
            )
        ax.legend()
        fig.savefig(filename, dpi=args.dpi, bbox_inches="tight")
    plt.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Optional

import jax
from jax import lax, numpy as jnp
from jax.scipy.stats import truncnorm
from jaxtyping import Array, ArrayLike
from numpyro.distributions import constraints, Distribution
from numpyro.distributions.util import promote_shapes, validate_sample

from ...utils.kernel import log_planck_taper_window
from ..constraints import mass_ratio_mass_sandwich
from ..utils import doubly_truncated_power_law_log_norm_constant


@jax.jit
def _broken_powerlaw_prob(
    m1: Array,
    alpha1: Array,
    alpha2: Array,
    mmin: Array,
    mmax: Array,
    mbreak: Array,
) -> Array:
    r"""Calculate the log probability of the broken powerlaw two peak distribution.

    The broken powerlaw two peak distribution is defined as

    .. math::
        p(m_1\mid \alpha_1, \alpha_2, m_{\mathrm{min}}, m_{\mathrm{max}}, m_{\mathrm{break}}) \propto
        \begin{cases}
            \left(\frac{m_1}{m_{\mathrm{break}}}\right)^{-\alpha_1}
            & m_{\mathrm{min}} \leq m_1 < m_{\text{break}} \\
            \left(\frac{m_1}{m_{\mathrm{break}}}\right)^{-\alpha_2}
            & m_{\text{break}} \leq m_1 \leq m_{\mathrm{max}} \\
            0 & \text{otherwise}
        \end{cases}
    """
    log_mbreak = jnp.log(mbreak)
    log_m1 = jnp.log(m1)
    log_unnormalized = jnp.where(
        m1 < mbreak,
        -alpha1 * (log_m1 - log_mbreak),
        -alpha2 * (log_m1 - log_mbreak),
    )
    log_norm = jnp.logaddexp(
        alpha1 * log_mbreak
        + doubly_truncated_power_law_log_norm_constant(-alpha1, mmin, mbreak),
        alpha2 * log_mbreak
        + doubly_truncated_power_law_log_norm_constant(-alpha2, mbreak, mmax),
    )
    return jnp.exp(log_unnormalized - log_norm)


class BrokenPowerlawTwoPeak(Distribution):
    r"""Broken Powerlaw + 2 Peak is defined as a mixture of one Broken Powerlaw and two
    left truncated Normal distributions. For more details, see appendix B.3 of
    `GWTC-4.0: Population Properties of Merging Compact Binaries <https://arxiv.org/abs/2508.18083>`_.

    It is defined as follows:

    .. math::

        p_{\mathrm{BP}}(m_1 \mid \alpha_1, \alpha_2, m_{1,\mathrm{min}}, m_{\mathrm{max}},
        m_{\mathrm{break}}) \propto
        \begin{cases}
            \left(\frac{m_1}{m_{\mathrm{break}}}\right)^{-\alpha_1}
            & m_{1,\mathrm{min}} \leq m_1 < m_{\text{break}}  \\
            \left(\frac{m_1}{m_{\mathrm{break}}}\right)^{-\alpha_2}
            & m_{\text{break}} \leq m_1 \leq m_{\mathrm{max}} \\
            0 & \text{otherwise}
        \end{cases}

    .. math::

        p(m_1 \mid \alpha_1, \alpha_2, m_{1,\mathrm{min}}, m_{\mathrm{max}},
        m_{\mathrm{break}}, \lambda_0, \lambda_1, \mu_1, \sigma_1, \mu_2, \sigma_2)
        \propto
        \left(
        \lambda_0 p_{\mathrm{BP}}(m_1 \mid \alpha_1, \alpha_2, m_{1,\mathrm{min}}, m_{\mathrm{max}}, m_{\mathrm{break}}) +
        \lambda_1 \mathcal{N}_{[m_{1,\mathrm{min}}, \infty)}(m_1 \mid \mu_1, \sigma_1) +
        (1 - \lambda_0 - \lambda_1) \mathcal{N}_{[m_{1,\mathrm{min}}, \infty)}(m_1 \mid \mu_2, \sigma_2)
        \right)
        S\left(\frac{m_1 - m_{1,\mathrm{min}}}{\delta_{m_1}}\right)

    .. math::

        p(q \mid \beta, m_1, m_{2,\mathrm{min}}, \delta_{m_2}) \propto
        q^{\beta} S\left(\frac{m_1 q - m_{2,\mathrm{min}}}{\delta_{m_2}}\right)

    .. math::
        p(m_1, q \mid \alpha_1, \alpha_2, m_{1,\mathrm{min}}, m_{2,\mathrm{min}},
        m_{\mathrm{max}}, m_{\mathrm{break}}, \lambda_0, \lambda_1, \mu_1, \sigma_1,
        \mu_2, \sigma_2, \beta, \delta_{m_1}, \delta_{m_2}) =
        p(m_1 \mid \alpha_1, \alpha_2, m_{1,\mathrm{min}}, m_{\mathrm{max}}, m_{\mathrm{break}}, \lambda_0,
        \lambda_1, \mu_1, \sigma_1, \mu_2, \sigma_2)
        p(q \mid \beta, m_1, m_{2,\mathrm{min}}, \delta_{m_2})

    where :math:`\mathcal{N}_{[m_{1,\mathrm{min}}, \infty)}(m \mid \mu, \sigma)` is a left
    truncated normal distribution with mean :math:`\mu` and standard deviation
    :math:`\sigma`, :math:`\lambda_0` and :math:`\lambda_1` are the mixing fractions of
    the broken powerlaw and first Gaussian component respectively, :math:`\delta_{m_1}`
    is the smoothing scale for the primary mass, and :math:`S(x)` is the exponential of
    :func:`~gwkokab.utils.kernel.log_planck_taper_window`.
    """

    arg_constraints = {
        "alpha1": constraints.real,
        "alpha2": constraints.real,
        "beta": constraints.real,
        "delta_m1": constraints.positive,
        "delta_m2": constraints.positive,
        "lambda_0": constraints.unit_interval,
        "lambda_1": constraints.unit_interval,
        "loc1": constraints.positive,
        "loc2": constraints.positive,
        "m1min": constraints.real,
        "m2min": constraints.real,
        "mbreak": constraints.real,
        "mmax": constraints.real,
        "scale1": constraints.positive,
        "scale2": constraints.positive,
    }

    pytree_data_fields = (
        "_logZ",
        "_m1s",
        "_support",
        "_Z_q_given_m1",
        "alpha1",
        "alpha2",
        "beta",
        "delta_m1",
        "delta_m2",
        "lambda_0",
        "lambda_1",
        "loc1",
        "loc2",
        "m1min",
        "m2min",
        "mbreak",
        "mmax",
        "scale1",
        "scale2",
    )

    def __init__(
        self,
        alpha1: ArrayLike,
        alpha2: ArrayLike,
        beta: ArrayLike,
        loc1: ArrayLike,
        loc2: ArrayLike,
        scale1: ArrayLike,
        scale2: ArrayLike,
        delta_m1: ArrayLike,
        delta_m2: ArrayLike,
        lambda_0: ArrayLike,
        lambda_1: ArrayLike,
        m1min: ArrayLike,
        m2min: ArrayLike,
        mmax: ArrayLike,
        mbreak: ArrayLike,
        validate_args: Optional[bool] = None,
    ) -> None:
        (
            self.alpha1,
            self.alpha2,
            self.beta,
            self.loc1,
            self.loc2,
            self.scale1,
            self.scale2,
            self.delta_m1,
            self.delta_m2,
            self.lambda_0,
            self.lambda_1,
            self.m1min,
            self.m2min,
            self.mmax,
            self.mbreak,
        ) = promote_shapes(
            alpha1,
            alpha2,
            beta,
            loc1,
            loc2,
            scale1,
            scale2,
            delta_m1,
            delta_m2,
            lambda_0,
            lambda_1,
            m1min,
            m2min,
            mmax,
            mbreak,
        )
        batch_shape = lax.broadcast_shapes(
            jnp.shape(alpha1),
            jnp.shape(alpha2),
            jnp.shape(beta),
            jnp.shape(loc1),
            jnp.shape(loc2),
            jnp.shape(scale1),
            jnp.shape(scale2),
            jnp.shape(delta_m1),
            jnp.shape(delta_m2),
            jnp.shape(lambda_0),
            jnp.shape(lambda_1),
            jnp.shape(m1min),
            jnp.shape(m2min),
            jnp.shape(mmax),
            jnp.shape(mbreak),
        )

        self._support = mass_ratio_mass_sandwich(m2min, mmax)
        super(BrokenPowerlawTwoPeak, self).__init__(
            batch_shape=batch_shape, event_shape=(2,), validate_args=validate_args
        )

        m1min = jnp.broadcast_to(m1min, batch_shape)
        m2min = jnp.broadcast_to(m2min, batch_shape)
        mmax = jnp.broadcast_to(mmax, batch_shape)

        m1s = jnp.linspace(m1min, mmax, 1000)
        self._logZ = jnp.log(
            jnp.trapezoid(
                jnp.exp(self._log_prob_m1_unnorm(m1s)),
                m1s,
                axis=0,
            )
        )

        self._m1s = jnp.linspace(m2min, mmax, 1000)
        _qs = jnp.linspace(0.005, 1.0, 500)
        _m1qs_grid = jnp.stack(jnp.meshgrid(self._m1s, _qs, indexing="ij"), axis=-1)

        _prob_q = jnp.exp(self._log_prob_q_unnorm(_m1qs_grid))

        self._Z_q_given_m1 = jnp.trapezoid(_prob_q, _qs, axis=1)

    @constraints.dependent_property(is_discrete=False, event_dim=1)
    def support(self) -> constraints.Constraint:
        return self._support

    def _log_prob_m1_unnorm(self, m1: Array) -> Array:
        safe_delta = jnp.where(self.delta_m1 <= 0.0, 1.0, self.delta_m1)
        log_smoothing_m1 = log_planck_taper_window((m1 - self.m1min) / safe_delta)
        broken_powerlaw_prob = _broken_powerlaw_prob(
            m1, self.alpha1, self.alpha2, self.m1min, self.mmax, self.mbreak
        )
        prob_norm_0 = truncnorm.pdf(
            m1,
            a=(self.m1min - self.loc1) / self.scale1,
            b=(self.mmax - self.loc1) / self.scale1,
            loc=self.loc1,
            scale=self.scale1,
        )
        prob_norm_1 = truncnorm.pdf(
            m1,
            a=(self.m1min - self.loc2) / self.scale2,
            b=(self.mmax - self.loc2) / self.scale2,
            loc=self.loc2,
            scale=self.scale2,
        )
        lambda_2 = 1.0 - self.lambda_0 - self.lambda_1

        log_prob_m1 = (
            jnp.log(
                self.lambda_0 * broken_powerlaw_prob
                + self.lambda_1 * prob_norm_0
                + lambda_2 * prob_norm_1
            )
            + log_smoothing_m1
        )

        return jnp.where(
            (self.delta_m1 <= 0.0) | (m1 < self.m1min), -jnp.inf, log_prob_m1
        )

    @validate_sample
    def _log_prob_q_unnorm(self, value: Array) -> Array:
        m1, q = jnp.unstack(value, axis=-1)
        m2 = m1 * q
        safe_delta = jnp.where(self.delta_m2 <= 0.0, 1.0, self.delta_m2)
        log_smoothing_q = log_planck_taper_window((m2 - self.m2min) / safe_delta)
        log_prob_q = self.beta * jnp.log(q) + log_smoothing_q

        return jnp.where(
            (self.delta_m2 <= 0.0) | (m2 < self.m2min), -jnp.inf, log_prob_q
        )

    @validate_sample
    def log_prob(self, value: ArrayLike) -> ArrayLike:
        m1, _ = jnp.unstack(value, axis=-1)
        log_prob_m1 = self._log_prob_m1_unnorm(m1) - self._logZ
        _Z_q = jnp.interp(m1, self._m1s, self._Z_q_given_m1, left=1.0, right=1.0)
        safe_Z_q = jnp.where(_Z_q <= 0, 1.0, _Z_q)
        log_Z_q = jnp.where(_Z_q <= 0, 0.0, jnp.log(safe_Z_q))
        log_prob_q = self._log_prob_q_unnorm(value) - log_Z_q
        return log_prob_m1 + log_prob_q



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import json
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd


def read_json(json_file: str) -> Dict:
    """Read json file and return.

    Parameters
    ----------
    json_file : str
        path of the json file

    Returns
    -------
    dict
        json file content as dict

    Raises
    ------
    ValueError
        If the file is not found or if the file is not a valid json file
    """
    try:
        with open(json_file, "r") as f:
            content = json.load(f)
    except (FileNotFoundError, json.JSONDecodeError) as e:
        raise ValueError(f"Error loading configuration: {e}")
    return content


def write_json(json_file: str, content: Dict) -> None:
    """Write a dictionary to a json file.

    Parameters
    ----------
    json_file : str
        path of the json file
    content : dict
        content to write to the json file

    Raises
    ------
    ValueError
        If the file is not writable or if the content is not a valid json serializable object
    """
    try:
        with open(json_file, "w") as f:
            json.dump(content, f, indent=4)
    except (FileNotFoundError, TypeError) as e:
        raise ValueError(f"Error writing configuration: {e}")


def expand_arguments(arg: str, n: int) -> List[str]:
    """Extend the argument with a number of strings.

    .. code:: python

        >>> expand_arguments("physics", 3)
        ["physics_0", "physics_1", "physics_2"]

    Parameters
    ----------
    arg : str
        argument to extend
    n : int
        number of strings to extend

    Returns
    -------
    List[str]
        list of extended arguments
    """
    return [f"{arg}_{i}" for i in range(n)]


def get_posterior_data(
    filenames: List[str], posterior_columns: List[str]
) -> List[np.ndarray]:
    """Get the posterior data from a list of files.

    Parameters
    ----------
    filenames : List[str]
        list of filenames
    posterior_columns : List[str]
        list of posterior columns

    Returns
    -------
    List[np.ndarray]
        dictionary of posterior data

    Raises
    ------
    ValueError
        If no files are found to read posterior data
    KeyError
        If the file is missing required columns
    """
    if len(filenames) == 0:
        raise ValueError("No files found to read posterior data")
    data_list = []
    for event in filenames:
        df = pd.read_csv(event, delimiter=" ")
        missing_columns = set(posterior_columns) - set(df.columns)
        if missing_columns:
            raise KeyError(
                f"The file '{event}' is missing required columns: {missing_columns}"
            )
        data = df[posterior_columns].to_numpy()
        data_list.append(data)
    return data_list


def ppd_ranges(
    parameters: List[str], ranges: List[Tuple[str, float, float, int]]
) -> List[Tuple[float, float, int]]:
    """Convert the PPD ranges to the format required by the PPD function.

    Parameters
    ----------
    parameters : List[str]
        list of parameters
    ranges : List[Tuple[str, float, float, int]]
        list of ranges

    Returns
    -------
    List[Tuple[float, float, int]]
        list of ranges

    Raises
    ------
    ValueError
        If the parameter is not found in the list of parameters
    """
    _ranges: List[Tuple[float, float, int]] = [None] * len(parameters)
    for name, *_range in ranges:
        if name not in parameters:
            raise ValueError(f"Parameter {name} not found in {parameters}.")
        _ranges[parameters.index(name)] = (
            float(_range[0]),
            float(_range[1]),
            int(_range[2]),
        )
    return _ranges



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import json
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import List

from jax import numpy as jnp, random as jrd
from loguru import logger
from numpyro import distributions as dist

from gwkokab.errors import banana_error_m1_m2
from gwkokab.parameters import Parameters as P
from gwkokab.poisson_mean import get_selection_fn_and_poisson_mean_estimator
from kokab.core.population import error_magazine, PopulationFactory
from kokab.ecc_matters.common import EccentricityMattersModel
from kokab.utils import genie_parser
from kokab.utils.common import read_json
from kokab.utils.logger import log_info
from kokab.utils.regex import match_all


def make_parser() -> ArgumentParser:
    """Create the command line argument parser.

    This function creates the command line argument parser and returns it.

    Returns
    -------
    ArgumentParser
        the command line argument parser
    """
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = genie_parser.get_parser(parser)
    parser.description = "Generate a population of CBCs"
    parser.epilog = "This script generates a population of CBCs"

    model_group = parser.add_argument_group("Model Options")

    model_group.add_argument(
        "--model-json",
        help="Path to the JSON file containing the model parameters",
        type=str,
        required=True,
    )

    err_group = parser.add_argument_group("Error Options")
    err_group.add_argument(
        "--err-json",
        help="Path to the JSON file containing the error parameters",
        type=str,
        required=True,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    log_info(start=True)

    with open(args.model_json, "r") as f:
        model_json = json.load(f)

    with open(args.err_json, "r") as f:
        err_json = json.load(f)

    model_params_name: List[str] = [
        "alpha_m",
        "high",
        "loc",
        "log_rate",
        "low",
        "mmax",
        "mmin",
        "scale",
    ]

    model_param = match_all(model_params_name, model_json)
    err_param = match_all(
        ["scale_Mc", "scale_eta", "loc", "scale", "low", "high"], err_json
    )

    error_magazine.register(
        (P.PRIMARY_MASS_SOURCE.value, P.SECONDARY_MASS_SOURCE.value),
        lambda x, size, key: banana_error_m1_m2(
            x,
            size,
            key,
            scale_Mc=err_param["scale_Mc"],
            scale_eta=err_param["scale_eta"],
        ),
    )

    @error_magazine.register(P.ECCENTRICITY.value)
    def ecc_error_fn(x, size, key):
        err_x = dist.TruncatedNormal(
            loc=x,
            scale=err_param["scale"],
            low=err_param["low"],
            high=err_param["high"],
        ).sample(key=key, sample_shape=(size,))
        mask = err_x < 0.0
        mask |= err_x > 1.0
        err_x = jnp.where(mask, jnp.full_like(mask, jnp.nan), err_x)
        return err_x

    parameters_name = [
        P.PRIMARY_MASS_SOURCE.value,
        P.SECONDARY_MASS_SOURCE.value,
        P.ECCENTRICITY.value,
    ]

    logger.info(f"Setting the random number generator key with seed {args.seed}.")
    pmean_key, factory_key = jrd.split(jrd.PRNGKey(args.seed), 2)

    pmean_config = read_json(args.pmean_json)
    log_selection_fn, erate_estimator, _, _ = (
        get_selection_fn_and_poisson_mean_estimator(
            key=pmean_key, parameters=parameters_name, **pmean_config
        )
    )

    popfactory = PopulationFactory(
        model_fn=EccentricityMattersModel,
        model_params=model_param,
        parameters=parameters_name,
        log_selection_fn=log_selection_fn,
        poisson_mean_estimator=erate_estimator,
        num_realizations=args.num_realizations,
        error_size=args.error_size,
        tile_covariance=args.tile_covariance,
    )

    popfactory.produce(factory_key)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Callable, Sequence
from typing import List, Optional, Tuple

import equinox as eqx
import h5py
import jax
import numpy as np
from jax import nn as jnn, numpy as jnp
from jaxtyping import Array, PRNGKeyArray
from loguru import logger

from ..constants import SECONDS_PER_YEAR
from ..models.utils import ScaledMixture
from ..parameters import Parameters as P
from ..utils.tools import batch_and_remainder, error_if, warn_if


_PARAM_MAPPING = {
    P.PRIMARY_MASS_SOURCE.value: "mass1_source",
    P.PRIMARY_SPIN_X.value: "spin1x",
    P.PRIMARY_SPIN_Y.value: "spin1y",
    P.PRIMARY_SPIN_Z.value: "spin1z",
    P.REDSHIFT.value: "redshift",
    P.SECONDARY_MASS_SOURCE.value: "mass2_source",
    P.SECONDARY_SPIN_X.value: "spin2x",
    P.SECONDARY_SPIN_Y.value: "spin2y",
    P.SECONDARY_SPIN_Z.value: "spin2z",
}


def load_o1o2o3_or_endO_injection_data(
    filename: str,
    parameters: Sequence[str],
    far_cut: float = 1.0,
    snr_cut: float = 10.0,
    ifar_pipelines: Sequence[str] | None = None,
):
    if P.ECCENTRICITY.value in parameters:
        _PARAM_MAPPING[P.ECCENTRICITY.value] = P.ECCENTRICITY.value
        warn_if(
            True,
            msg="Eccentricity injections are not part of O1, O2 or O3 injections. "
            f"Make sure you have altered the injection set with {P.ECCENTRICITY.value} "
            "key accordingly.",
        )
    with h5py.File(filename, "r") as f:
        analysis_time_years = float(f.attrs["analysis_time_s"]) / SECONDS_PER_YEAR
        logger.debug("Analysis time: {:.2f} years", analysis_time_years)

        total_injections = int(f.attrs["total_generated"])
        logger.debug("Total injections: {}", total_injections)

        injections = f["injections"]

        if ifar_pipelines is None:
            ifar_pipelines = [k for k in injections.keys() if "ifar" in k]
            logger.debug(
                "No pipelines specified for ifar, using all available: {}",
                ", ".join(ifar_pipelines),
            )
        else:
            logger.debug("Selecting ifar from pipelines: {}", ", ".join(ifar_pipelines))

        ifar = np.max([injections[k][:] for k in ifar_pipelines], axis=0)
        snr = injections["optimal_snr_net"][:]

        runs = injections["name"][:].astype(str)
        found = np.where(runs == "o3", ifar > 1 / far_cut, snr > snr_cut)

        n_total = found.shape[0]
        n_found = np.sum(found)
        logger.debug(
            "Found {} out of {} injections with FAR < {} and SNR > {}",
            n_found,
            n_total,
            far_cut,
            snr_cut,
        )

        sampling_prob = (
            injections["sampling_pdf"][found][:]
            / injections["mixture_weight"][found][:]
        )

        χ_1x = injections["spin1x"][found][:]
        χ_1y = injections["spin1y"][found][:]
        χ_1z = injections["spin1z"][found][:]
        χ_2x = injections["spin2x"][found][:]
        χ_2y = injections["spin2y"][found][:]
        χ_2z = injections["spin2z"][found][:]
        a1 = np.sqrt(np.square(χ_1x) + np.square(χ_1y) + np.square(χ_1z))
        a2 = np.sqrt(np.square(χ_2x) + np.square(χ_2y) + np.square(χ_2z))

        injs = []
        for p in parameters:
            if p == P.COS_TILT_1.value:
                _inj = χ_1z / a1
            elif p == P.COS_TILT_2.value:
                _inj = χ_2z / a2
            elif p == P.PRIMARY_SPIN_MAGNITUDE.value:
                _inj = a1
                # We parameterize spins in spherical coordinates, neglecting azimuthal
                # P. The injections are parameterized in terms of cartesian
                # spins. The Jacobian is `1 / (2 pi magnitude ** 2)`.
                logger.debug(
                    "Correcting sampling probability for spherical spin "
                    "parameterization of primary spin."
                )
                sampling_prob *= 2.0 * np.pi * np.square(a1)
            elif p == P.SECONDARY_SPIN_MAGNITUDE.value:
                _inj = a2
                # We parameterize spins in spherical coordinates, neglecting azimuthal
                # P. The injections are parameterized in terms of cartesian
                # spins. The Jacobian is `1 / (2 pi magnitude ** 2)`.
                logger.debug(
                    "Correcting sampling probability for spherical spin "
                    "parameterization of secondary spin."
                )
                sampling_prob *= 2.0 * np.pi * np.square(a2)
            else:
                _inj = injections[_PARAM_MAPPING[p]][found][:]
            injs.append(_inj)

        if (
            P.PRIMARY_SPIN_MAGNITUDE.value not in parameters
            and P.SECONDARY_SPIN_MAGNITUDE.value not in parameters
        ):
            # Eliminating the probability of cartesian spins
            logger.debug(
                "Eliminating the probability of cartesian spins from sampling probability."
            )
            sampling_prob *= np.square(4.0 * np.pi * a1 * a2)

        injections = jax.device_put(np.stack(injs, axis=-1))

        sampling_log_prob = jax.device_put(np.log(sampling_prob))

    return injections, sampling_log_prob, analysis_time_years, total_injections


def load_o1o2o3o4a_injection_data(
    filename: str,
    parameters: Sequence[str],
    far_cut: float = 1.0,
    snr_cut: float = 10.0,
):
    expected_params = {
        P.COS_TILT_1.value,
        P.COS_TILT_2.value,
        P.PRIMARY_MASS_SOURCE.value,
        P.PRIMARY_SPIN_MAGNITUDE.value,
        P.REDSHIFT.value,
        P.SECONDARY_MASS_SOURCE.value,
        P.SECONDARY_SPIN_MAGNITUDE.value,
    }
    if P.ECCENTRICITY.value in parameters:
        _PARAM_MAPPING[P.ECCENTRICITY.value] = P.ECCENTRICITY.value
        expected_params.add(P.ECCENTRICITY.value)
        warn_if(
            True,
            msg="Eccentricity injections are not part of O1, O2, O3, or O4a injections. "
            f"Make sure you have altered the injection set with {P.ECCENTRICITY.value} "
            "key accordingly.",
        )

    error_if(
        set(parameters) != expected_params,
        msg=(
            "For O1, O2, O3, and O4a injections, the following parameters must be used: "
            + ", ".join(_PARAM_MAPPING.keys())
        ),
    )

    with h5py.File(filename, "r") as f:
        raw_events = f["events"]
        keys: List[str] = list(raw_events.dtype.names)

        analysis_time_years = float(f.attrs["total_analysis_time"]) / SECONDS_PER_YEAR
        logger.debug("Analysis time: {:.2f} years", analysis_time_years)

        total_injections = int(f.attrs["total_generated"])
        logger.debug("Total injections: {}", total_injections)

        far_pipelines = [
            k
            for k in keys
            if any([k.startswith("far_"), k.endswith("_far"), "_far_" in k])
        ]

        ifar_pipelines = [
            k
            for k in keys
            if any([k.startswith("ifar_"), k.endswith("_ifar"), "_ifar_" in k])
        ]

        if len(far_pipelines) > 0:
            logger.debug("Selecting far from pipelines: {}", ", ".join(far_pipelines))
        if len(ifar_pipelines) > 0:
            logger.debug("Selecting ifar from pipelines: {}", ", ".join(ifar_pipelines))

        ifar = np.max(
            [1.0 / raw_events[k][:] for k in far_pipelines]
            + [raw_events[k][:] for k in ifar_pipelines],
            axis=0,
        )
        snr = raw_events["semianalytic_observed_phase_maximized_snr_net"][:]

        found = (snr > snr_cut) | (ifar > 1 / far_cut)

        n_total = int(f.attrs["num_accepted"])
        n_found = np.sum(found)
        logger.debug(
            "Found {} out of {} injections with FAR < {} and SNR > {}",
            n_found,
            n_total,
            far_cut,
            snr_cut,
        )

        χ_1x = raw_events["spin1x"][found][:]
        χ_1y = raw_events["spin1y"][found][:]
        χ_1z = raw_events["spin1z"][found][:]
        χ_2x = raw_events["spin2x"][found][:]
        χ_2y = raw_events["spin2y"][found][:]
        χ_2z = raw_events["spin2z"][found][:]
        a1 = np.sqrt(np.square(χ_1x) + np.square(χ_1y) + np.square(χ_1z))
        a2 = np.sqrt(np.square(χ_2x) + np.square(χ_2y) + np.square(χ_2z))

        sampling_log_prob = (
            raw_events[
                "lnpdraw_mass1_source_mass2_source_redshift_spin1x_spin1y_spin1z_spin2x_spin2y_spin2z"
            ][found][:]
            - np.log(raw_events["weights"][found][:])
            # We parameterize spins in spherical coordinates, neglecting azimuthal
            # P. The injections are parameterized in terms of cartesian
            # spins. The Jacobian is `1 / (2 pi magnitude ** 2)`.
            + 2.0 * (np.log(2.0 * np.pi) + np.log(a1) + np.log(a2))
        )

        injs = []
        for p in parameters:
            if p == P.COS_TILT_1.value:
                _inj = χ_1z / a1
            elif p == P.COS_TILT_2.value:
                _inj = χ_2z / a2
            elif p == P.PRIMARY_SPIN_MAGNITUDE.value:
                _inj = a1
            elif p == P.SECONDARY_SPIN_MAGNITUDE.value:
                _inj = a2
            else:
                _inj = raw_events[_PARAM_MAPPING[p]][found][:]
            injs.append(_inj)

        events = jax.device_put(np.stack(injs, axis=-1))

        sampling_log_prob = jax.device_put(sampling_log_prob)

    return events, sampling_log_prob, analysis_time_years, total_injections


def poisson_mean_from_sensitivity_injections(
    key: PRNGKeyArray,
    parameters: Sequence[str],
    filename: str,
    batch_size: Optional[int] = None,
    far_cut: float = 1.0,
    snr_cut: float = 10.0,
    ifar_pipelines: Sequence[str] | None = None,
) -> Tuple[
    Optional[Callable[[Array], Array]],
    Callable[[ScaledMixture], Array],
    float | Array,
    Callable[[ScaledMixture], Array],
]:
    del key  # Unused.

    with h5py.File(filename, "r") as f:
        is_o1o2o3o4a = "events" in f

    if is_o1o2o3o4a:
        del ifar_pipelines  # Unused.
        # θ_i, log w_i, T, N_total
        samples, log_weights, analysis_time_years, total_injections = (
            load_o1o2o3o4a_injection_data(
                filename,
                parameters,
                far_cut,
                snr_cut,
            )
        )
    else:
        # θ_i, log w_i, T, N_total
        samples, log_weights, analysis_time_years, total_injections = (
            load_o1o2o3_or_endO_injection_data(
                filename,
                parameters,
                far_cut,
                snr_cut,
                ifar_pipelines,
            )
        )

    n_accepted = log_weights.shape[0]

    def _poisson_mean(scaled_mixture: ScaledMixture) -> Array:
        log_prob_fn = eqx.filter_jit(eqx.filter_vmap(scaled_mixture.log_prob))

        def _f(carry_logsumexp: Array, data: Tuple[Array, Array]) -> Tuple[Array, None]:
            log_weights, samples = data
            # log p(θ_i|λ)
            model_log_prob = log_prob_fn(samples).reshape(log_weights.shape[0])
            # log p(θ_i|λ) - log w_i
            log_prob = model_log_prob - log_weights
            safe_log_prob = jnp.where(
                jnp.isneginf(log_prob) | jnp.isnan(log_prob),
                -jnp.inf,
                log_prob,
            )

            partial_logsumexp = jnn.logsumexp(
                safe_log_prob,
                where=~jnp.isneginf(safe_log_prob),
                axis=-1,
            )
            safe_carry_logsumexp = jnp.where(
                jnp.isneginf(carry_logsumexp) | jnp.isnan(carry_logsumexp),
                -jnp.inf,
                carry_logsumexp,
            )
            return jnp.logaddexp(safe_carry_logsumexp, partial_logsumexp), None

        initial_logprob = jnp.asarray(-jnp.inf)
        if batch_size is None or n_accepted <= batch_size:
            # If the number of accepted injections is less than or equal to the batch size,
            # we can process them all at once.
            log_prob, _ = _f(initial_logprob, (log_weights, samples))
        else:
            batched_log_weights, remainder_log_weights = batch_and_remainder(
                log_weights, batch_size
            )
            batched_samples, remainder_samples = batch_and_remainder(
                samples, batch_size
            )
            batched_logprob, _ = jax.lax.scan(
                _f,
                initial_logprob,
                (batched_log_weights, batched_samples),
            )
            log_prob, _ = _f(
                batched_logprob, (remainder_log_weights, remainder_samples)
            )

        # (T / n_total) * exp(log Σ exp(log p(θ_i|λ) - log w_i))
        return (analysis_time_years / total_injections) * jnp.exp(log_prob)

    @eqx.filter_jit
    def _variance_of_estimator(scaled_mixture: ScaledMixture) -> Array:
        """See equation 9 and 11 of https://arxiv.org/abs/2406.16813."""
        log_prob_fn = eqx.filter_jit(eqx.filter_vmap(scaled_mixture.log_prob))

        def _f(
            carry: Tuple[Array, Array], data: Tuple[Array, Array]
        ) -> Tuple[Tuple[Array, Array], None]:
            carry_logsumexp, carry_logsumexp2 = carry
            log_weights, samples = data
            # log p(θ_i|λ)
            model_log_prob = log_prob_fn(samples).reshape(log_weights.shape[0])
            # log p(θ_i|λ) - log w_i
            log_prob = model_log_prob - log_weights
            safe_log_prob = jnp.where(
                jnp.isneginf(log_prob) | jnp.isnan(log_prob),
                -jnp.inf,
                log_prob,
            )

            partial_logsumexp = jnn.logsumexp(
                safe_log_prob,
                where=~jnp.isneginf(safe_log_prob),
                axis=-1,
            )
            partial_logsumexp2 = jnn.logsumexp(
                2.0 * safe_log_prob,
                where=~jnp.isneginf(safe_log_prob),
                axis=-1,
            )

            safe_carry_logsumexp = jnp.where(
                jnp.isneginf(carry_logsumexp) | jnp.isnan(carry_logsumexp),
                -jnp.inf,
                carry_logsumexp,
            )
            safe_carry_logsumexp2 = jnp.where(
                jnp.isneginf(carry_logsumexp2) | jnp.isnan(carry_logsumexp2),
                -jnp.inf,
                carry_logsumexp2,
            )

            return (
                jnp.logaddexp(safe_carry_logsumexp, partial_logsumexp),
                jnp.logaddexp(safe_carry_logsumexp2, partial_logsumexp2),
            ), None

        initial_logprob = jnp.asarray(-jnp.inf)
        if batch_size is None or n_accepted <= batch_size:
            # If the number of accepted injections is less than or equal to the batch size,
            # we can process them all at once.
            (log_prob, log_prob2), _ = _f(
                (initial_logprob, initial_logprob), (log_weights, samples)
            )
        else:
            batched_log_weights, remainder_log_weights = batch_and_remainder(
                log_weights, batch_size
            )
            batched_samples, remainder_samples = batch_and_remainder(
                samples, batch_size
            )
            (batched_logprob, batched_logprob2), _ = jax.lax.scan(
                _f,
                (initial_logprob, initial_logprob),
                (batched_log_weights, batched_samples),
            )
            (log_prob, log_prob2), _ = _f(
                (batched_logprob, batched_logprob2),
                (remainder_log_weights, remainder_samples),
            )

        term2 = jnp.exp(
            2.0 * jnp.log(analysis_time_years)
            - 3.0 * jnp.log(total_injections)
            + 2.0 * log_prob
        )
        term1 = jnp.exp(
            2.0 * jnp.log(analysis_time_years)
            - 2.0 * jnp.log(total_injections)
            + log_prob2
        )
        return term1 - term2

    return None, _poisson_mean, analysis_time_years, _variance_of_estimator



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

from .poisson_factory import (
    get_selection_fn_and_poisson_mean_estimator as get_selection_fn_and_poisson_mean_estimator,
)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Optional

from jax import numpy as jnp
from jaxtyping import Array
from numpyro.distributions import HalfNormal, Independent, TruncatedNormal

from ..mass import BrokenPowerlawTwoPeak
from ..redshift import PowerlawRedshift
from ..spin import MinimumTiltModel
from ..transformations import PrimaryMassAndMassRatioToComponentMassesTransform
from ..utils import (
    ExtendedSupportTransformedDistribution,
    JointDistribution,
    ScaledMixture,
)


def BrokenPowerlawTwoPeakFull(
    use_spin: bool = False,
    use_redshift: bool = False,
    use_tilt: bool = False,
    use_eccentricity: bool = False,
    validate_args: Optional[bool] = None,
    **params: Array,
) -> ScaledMixture:
    # NOTE: If you change something here, please also change in
    # kokab/bp2pfull/ppd.py
    smoothing_model = ExtendedSupportTransformedDistribution(
        BrokenPowerlawTwoPeak(
            alpha1=params["alpha1"],
            alpha2=params["alpha2"],
            beta=params["beta"],
            loc1=params["loc1"],
            loc2=params["loc2"],
            scale1=params["scale1"],
            scale2=params["scale2"],
            delta_m1=params["delta_m1"],
            delta_m2=params["delta_m2"],
            lambda_0=params["lambda_0"],
            lambda_1=params["lambda_1"],
            m1min=params["m1min"],
            m2min=params["m2min"],
            mmax=params["mmax"],
            mbreak=params["mbreak"],
            validate_args=validate_args,
        ),
        transforms=PrimaryMassAndMassRatioToComponentMassesTransform(),
        validate_args=validate_args,
    )

    component_distributions = [smoothing_model]

    if use_spin:
        chi_dist = Independent(
            TruncatedNormal(
                loc=jnp.stack([params["chi_loc"], params["chi_loc"]], axis=-1),
                scale=jnp.stack([params["chi_scale"], params["chi_scale"]], axis=-1),
                low=0.0,
                high=1.0,
                validate_args=validate_args,
            ),
            reinterpreted_batch_ndims=1,
            validate_args=validate_args,
        )
        component_distributions.append(chi_dist)

    if use_tilt:
        tilt_dist = MinimumTiltModel(
            zeta=params["cos_tilt_zeta"],
            loc=params["cos_tilt_loc"],
            scale=params["cos_tilt_scale"],
            minimum=params.get("cos_tilt_minimum", -1.0),
            validate_args=validate_args,
        )
        component_distributions.append(tilt_dist)

    if use_eccentricity:
        ecc_dist = HalfNormal(
            scale=params["eccentricity_scale"], validate_args=validate_args
        )
        component_distributions.append(ecc_dist)

    if use_redshift:
        z_max = params["z_max"]
        kappa = params["kappa"]
        powerlaw_z = PowerlawRedshift(
            z_max=z_max, kappa=kappa, validate_args=validate_args
        )
        component_distributions.append(powerlaw_z)

    if len(component_distributions) > 1:
        component_distributions = [
            JointDistribution(*component_distributions, validate_args=validate_args)
        ]

    return ScaledMixture(
        log_scales=jnp.asarray([params["log_rate"]]),
        component_distributions=component_distributions,
        support=component_distributions[0]._support,  # type: ignore
        validate_args=validate_args,
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from ._bp2p import BrokenPowerlawTwoPeak as BrokenPowerlawTwoPeak
from ._models import (
    PowerlawPrimaryMassRatio as PowerlawPrimaryMassRatio,
    SmoothedTwoComponentPrimaryMassRatio as SmoothedTwoComponentPrimaryMassRatio,
    Wysocki2019MassModel as Wysocki2019MassModel,
)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Callable, Sequence
from typing import Optional, Tuple

from jaxtyping import Array, PRNGKeyArray

from ..models.utils import ScaledMixture
from ..utils.tools import error_if
from ._injection_based import (
    poisson_mean_from_sensitivity_injections as _poisson_mean_from_sensitivity_injections,
)
from ._neural_pdet import (
    poisson_mean_from_neural_pdet as _poisson_mean_from_neural_pdet,
)
from ._neural_vt import (
    poisson_mean_from_neural_vt as _poisson_mean_from_neural_vt,
)


def get_selection_fn_and_poisson_mean_estimator(
    estimator_type: str,
    key: PRNGKeyArray,
    parameters: Sequence[str],
    filename: str,
    batch_size: Optional[int] = None,
    **kwargs,
) -> Tuple[
    Optional[Callable[[Array], Array]],
    Callable[[ScaledMixture], Array],
    float | Array,
    Callable[[ScaledMixture], Array],
]:
    valid_estimator_types = ("injection", "neural_vt", "neural_pdet", "custom")
    error_if(
        estimator_type not in valid_estimator_types,
        msg="estimator_type must be one of " + ", ".join(valid_estimator_types),
    )
    if estimator_type == "injection":
        ifar_pipelines = kwargs.pop("ifar_pipelines", None)
        return _poisson_mean_from_sensitivity_injections(
            key=key,
            parameters=parameters,
            filename=filename,
            batch_size=batch_size,
            far_cut=kwargs.pop("far_cut", 1.0),
            snr_cut=kwargs.pop("snr_cut", 10.0),
            ifar_pipelines=ifar_pipelines,
        )
    elif estimator_type == "neural_vt":
        return _poisson_mean_from_neural_vt(
            key=key,
            parameters=parameters,
            filename=filename,
            batch_size=batch_size,
            num_samples=kwargs.pop("num_samples", 1_000),
            time_scale=kwargs.pop("time_scale", 1.0),
        )
    elif estimator_type == "neural_pdet":
        return _poisson_mean_from_neural_pdet(
            key=key,
            parameters=parameters,
            filename=filename,
            batch_size=batch_size,
            num_samples=kwargs.pop("num_samples", 1_000),
            time_scale=kwargs.pop("time_scale", 1.0),
        )
    elif estimator_type == "custom":
        error_if(
            kwargs.get("python_module_path", None) is None,
            msg="For custom estimator_type, 'python_module_path' must be provided.",
        )

        python_module_path: str = kwargs.pop("python_module_path")
        error_if(
            not isinstance(python_module_path, str),
            msg="'python_module_path' must be a string, got "
            + str(type(python_module_path)),
        )
        error_if(
            not python_module_path.endswith(".py"),
            msg="'python_module_path' must point to a .py file, got '"
            + python_module_path
            + "'.",
        )

        import importlib.util

        spec = importlib.util.spec_from_file_location(
            "custom_module", python_module_path
        )  # type: ignore
        custom_module = importlib.util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(custom_module)  # type: ignore

        error_if(
            not hasattr(custom_module, "custom_poisson_mean_estimator"),
            msg="The custom module must have a 'custom_poisson_mean_estimator' function.",
        )

        return custom_module.custom_poisson_mean_estimator(
            key, parameters, filename, batch_size=batch_size, **kwargs
        )

    else:
        raise ValueError(f"Unknown estimator_type: {estimator_type}")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import Callable, Dict, List, Optional, Union

from jaxtyping import Array, ArrayLike
from numpyro._typing import DistributionLike
from numpyro.distributions.distribution import enable_validation

from gwkokab.inference import flowMC_poisson_likelihood, numpyro_poisson_likelihood
from gwkokab.models import BrokenPowerlawTwoPeakFull
from gwkokab.models.utils import JointDistribution, ScaledMixture
from gwkokab.parameters import Parameters as P
from kokab.core.flowMC_based import flowMC_arg_parser, FlowMCBased
from kokab.core.numpyro_based import numpyro_arg_parser, NumpyroBased
from kokab.core.sage import Sage, sage_arg_parser
from kokab.utils.logger import log_info


class BrokenPowerlawTwoPeakFullCore(Sage):
    def __init__(
        self,
        has_spin: bool,
        has_tilt: bool,
        has_eccentricity: bool,
        has_redshift: bool,
        likelihood_fn: Callable[
            [
                Callable[..., DistributionLike],
                JointDistribution,
                Dict[str, DistributionLike],
                Dict[str, int],
                ArrayLike,
                Callable[[ScaledMixture], Array],
                Optional[List[Callable[..., Array]]],
                Dict[str, Array],
            ],
            Callable,
        ],
        posterior_regex: str,
        posterior_columns: List[str],
        seed: int,
        prior_filename: str,
        poisson_mean_filename: str,
        sampler_settings_filename: str,
        variance_cut_threshold: Optional[float],
        n_buckets: int,
        threshold: float,
        debug_nans: bool = False,
        profile_memory: bool = False,
        check_leaks: bool = False,
    ) -> None:
        self.has_spin = has_spin
        self.has_tilt = has_tilt
        self.has_eccentricity = has_eccentricity
        self.has_redshift = has_redshift

        super().__init__(
            likelihood_fn=likelihood_fn,
            model=BrokenPowerlawTwoPeakFull,
            posterior_regex=posterior_regex,
            posterior_columns=posterior_columns,
            seed=seed,
            prior_filename=prior_filename,
            poisson_mean_filename=poisson_mean_filename,
            sampler_settings_filename=sampler_settings_filename,
            variance_cut_threshold=variance_cut_threshold,
            analysis_name="one_powerlaw_one_peak",
            n_buckets=n_buckets,
            threshold=threshold,
            debug_nans=debug_nans,
            profile_memory=profile_memory,
            check_leaks=check_leaks,
            where_fns=None,
        )

    @property
    def constants(self) -> Dict[str, Union[int, float, bool]]:
        return {
            "use_spin": self.has_spin,
            "use_tilt": self.has_tilt,
            "use_eccentricity": self.has_eccentricity,
            "use_redshift": self.has_redshift,
        }

    @property
    def parameters(self) -> List[str]:
        names = [P.PRIMARY_MASS_SOURCE.value, P.SECONDARY_MASS_SOURCE.value]
        if self.has_spin:
            names.append(P.PRIMARY_SPIN_MAGNITUDE.value)
            names.append(P.SECONDARY_SPIN_MAGNITUDE.value)
        if self.has_tilt:
            names.extend([P.COS_TILT_1.value, P.COS_TILT_2.value])
        if self.has_eccentricity:
            names.append(P.ECCENTRICITY.value)
        if self.has_redshift:
            names.append(P.REDSHIFT.value)
        return names

    @property
    def model_parameters(self) -> List[str]:
        model_parameters = [
            "alpha1",
            "alpha2",
            "beta",
            "delta_m1",
            "delta_m2",
            "lambda_0",
            "lambda_1",
            "loc1",
            "loc2",
            "log_rate",
            "m1min",
            "m2min",
            "mbreak",
            "mmax",
            "scale1",
            "scale2",
        ]

        if self.has_spin:
            model_parameters.extend(["chi_loc", "chi_scale"])

        if self.has_tilt:
            model_parameters.extend(["cos_tilt_zeta", "cos_tilt_loc", "cos_tilt_scale"])

        if self.has_eccentricity:
            model_parameters.append("eccentricity_scale")

        if self.has_redshift:
            model_parameters.extend(["kappa", "z_max"])

        return model_parameters


class BrokenPowerlawTwoPeakFullFSage(BrokenPowerlawTwoPeakFullCore, FlowMCBased):
    pass


class BrokenPowerlawTwoPeakFullNSage(BrokenPowerlawTwoPeakFullCore, NumpyroBased):
    pass


def model_arg_parser(parser: ArgumentParser) -> ArgumentParser:
    model_group = parser.add_argument_group("Model Options")
    model_group.add_argument(
        "--add-spin",
        action="store_true",
        help="Include truncated normal spin parameters in the model.",
    )
    model_group.add_argument(
        "--add-tilt",
        action="store_true",
        help="Include tilt parameters in the model.",
    )
    model_group.add_argument(
        "--add-eccentricity",
        action="store_true",
        help="Include eccentricity parameter in the model",
    )
    model_group.add_argument(
        "--add-redshift",
        action="store_true",
        help="Include redshift parameter in the model",
    )
    return parser


def f_main() -> None:
    enable_validation()

    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = model_arg_parser(parser)
    parser = sage_arg_parser(parser)
    parser = flowMC_arg_parser(parser)

    args = parser.parse_args()

    log_info(start=True)

    BrokenPowerlawTwoPeakFullFSage(
        has_spin=args.add_spin,
        has_tilt=args.add_tilt,
        has_eccentricity=args.add_eccentricity,
        has_redshift=args.add_redshift,
        likelihood_fn=flowMC_poisson_likelihood,
        posterior_regex=args.posterior_regex,
        posterior_columns=args.posterior_columns,
        seed=args.seed,
        prior_filename=args.prior_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        variance_cut_threshold=args.variance_cut_threshold,
        n_buckets=args.n_buckets,
        threshold=args.threshold,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
    ).run()


def n_main() -> None:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = model_arg_parser(parser)
    parser = sage_arg_parser(parser)
    parser = numpyro_arg_parser(parser)

    args = parser.parse_args()

    log_info(start=True)

    BrokenPowerlawTwoPeakFullNSage(
        has_spin=args.add_spin,
        has_tilt=args.add_tilt,
        has_eccentricity=args.add_eccentricity,
        has_redshift=args.add_redshift,
        likelihood_fn=numpyro_poisson_likelihood,
        posterior_regex=args.posterior_regex,
        posterior_columns=args.posterior_columns,
        seed=args.seed,
        prior_filename=args.prior_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        variance_cut_threshold=args.variance_cut_threshold,
        n_buckets=args.n_buckets,
        threshold=args.threshold,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
    ).run()



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Sequence
from typing import Callable, List, Optional, Tuple

import equinox as eqx
import h5py
import jax
import matplotlib.pyplot as plt
import numpy as np
import optax
import pandas as pd
import tqdm
from jax import nn as jnn, numpy as jnp, random as jrd
from jaxtyping import Array, PRNGKeyArray, PyTree
from loguru import logger
from numpyro.util import is_prng_key


# ==============================
# Losses & Prediction
# ==============================


@eqx.filter_value_and_grad
@eqx.filter_jit
def mse_loss_fn(
    model: PyTree, x: Array, y: Array, batch_size: Optional[int] = 256
) -> Array:
    """Mean squared error loss."""
    y_pred = jax.lax.map(model, x, batch_size=batch_size)
    return jnp.mean(jnp.square(y - y_pred))


@eqx.filter_value_and_grad
@eqx.filter_jit
def bce_logits_loss_fn(
    model: PyTree,
    x: Array,
    y: Array,
    batch_size: Optional[int] = 256,
    eps: float = 1e-6,
) -> Array:
    """Binary cross-entropy with logits (numerically stable).

    Expects targets in [0,1]; clips to [eps, 1-eps].
    """
    logits = jax.lax.map(model, x, batch_size=batch_size)
    y = jnp.clip(y, eps, 1.0 - eps)
    loss = optax.sigmoid_binary_cross_entropy(logits=logits, labels=y)
    return jnp.mean(loss)


@eqx.filter_jit
def predict(model: PyTree, x: Array, batch_size: Optional[int] = 256) -> Array:
    """Predict outputs for inputs x."""
    return jax.lax.map(model, x, batch_size=batch_size)


# ==============================
# IO Helpers
# ==============================


def read_data(data_path: str, keys: Sequence[str]) -> pd.DataFrame:
    """Read dataset (HDF5) into a DataFrame with columns = keys."""
    with h5py.File(data_path, "r") as vt_file:
        df = pd.DataFrame(data={key: np.array(vt_file[key]).flatten() for key in keys})
    return df


def make_model(
    *,
    key: PRNGKeyArray,
    input_layer: int,
    output_layer: int,
    width_size: int,
    depth: int,
) -> eqx.nn.MLP:
    """Build an MLP with ReLU activations."""
    assert is_prng_key(key)
    return eqx.nn.MLP(
        in_size=input_layer,
        out_size=output_layer,
        width_size=width_size,
        depth=depth,
        activation=jnn.relu,
        key=key,
    )


def save_model(
    *,
    filepath: str,
    datafilepath: str,
    model: eqx.nn.MLP,
    names: Optional[Sequence[str]] = None,
    is_log: bool = False,
) -> None:
    """Persist model weights and metadata to HDF5 (backward-compatible format)."""
    if not filepath.endswith(".hdf5"):
        raise ValueError("Model save path must end with .hdf5")

    with h5py.File(datafilepath, "r") as f:
        attr = dict(f.attrs)

    model: eqx.nn.MLP = model._fun  # type: ignore

    with h5py.File(filepath, "w") as f:
        for key, value in attr.items():
            f.attrs[key] = value
        if names is not None:
            f.create_dataset("names", data=np.array(names, dtype="S"))
        f.create_dataset("in_size", data=model.in_size)  # type: ignore
        f.create_dataset("out_size", data=model.out_size)  # type: ignore
        f.create_dataset("width_size", data=model.width_size)  # type: ignore
        f.create_dataset("depth", data=model.depth)  # type: ignore
        f.attrs["is_log"] = is_log
        num_layers = len(model.layers)  # type: ignore
        for i in range(num_layers):
            layer_i = f.create_group(f"layer_{i}")
            layer_i.create_dataset(f"weight_{i}", data=model.layers[i].weight)  # type: ignore
            layer_i.create_dataset(f"bias_{i}", data=model.layers[i].bias)  # type: ignore


def load_model(filename: str) -> Tuple[List[str], eqx.nn.MLP]:
    """Load model and names from HDF5 (backward-compatible)."""
    with h5py.File(filename, "r") as f:
        names = f["names"][:].astype(str).tolist()
        in_size = int(f["in_size"][()])
        out_size = int(f["out_size"][()])
        width_size = int(f["width_size"][()])
        depth = int(f["depth"][()])

        new_model = eqx.nn.MLP(
            in_size=in_size,
            out_size=out_size,
            width_size=width_size,
            depth=depth,
            activation=jnn.relu,
            key=jrd.PRNGKey(0),
        )

        i = 0
        while f.get(f"layer_{i}"):
            layer_i = f[f"layer_{i}"]
            weight_i = jax.device_put(np.asarray(layer_i[f"weight_{i}"][:]))
            bias_i = jax.device_put(np.asarray(layer_i[f"bias_{i}"][:]))
            new_model = eqx.tree_at(lambda m: m.layers[i].weight, new_model, weight_i)
            new_model = eqx.tree_at(lambda m: m.layers[i].bias, new_model, bias_i)
            i += 1

    return names, new_model


# ==============================
# Train/Validation Split
# ==============================


def _train_test_data_split(
    X: Array,
    Y: Array,
    batch_size: int,
    test_size: float = 0.2,
    seed: Optional[int] = None,
) -> tuple[Array, Array, Array, Array]:
    """Seeded split aligned to batch_size for stable validation curves."""
    n = len(X)
    if seed is not None:
        rng = np.random.default_rng(seed)
        indices = rng.permutation(n)
    else:
        indices = np.random.permutation(n)

    split = int(n * (1 - test_size))
    split = (
        max(split, batch_size) // batch_size
    ) * batch_size  # keep multiple of batch_size
    split = (
        min(split, n - batch_size) if n >= 2 * batch_size else max(batch_size, split)
    )

    train_indices = indices[:split]
    test_indices = indices[split:]
    return X[train_indices], X[test_indices], Y[train_indices], Y[test_indices]


# ==============================
# Trainer
# ==============================


def train_regressor(
    *,
    input_keys: list[str],
    output_keys: list[str],
    width_size: int,
    depth: int,
    batch_size: int,
    data_path: str,
    checkpoint_path: Optional[str] = None,
    epochs: int = 50,
    validation_split: float = 0.2,
    learning_rate: float = 1e-3,
    train_in_log: bool = False,
    # New, optional stabilizers (keep old calls working):
    loss_type: str = "mse",  # "mse" (Mean Squared Error) or "bce_logits" (Binary Cross-Entropy with logits)
    grad_clip_norm: float = 1.0,
    weight_decay: float = 1e-4,
    use_cosine_decay: bool = True,
    min_lr: float = 1e-6,
    warmup_epochs: int = 3,
    seed: Optional[int] = 42,
) -> None:
    """Train an MLP regressor with stable optimization and smooth loss curves.

    Notes
    -----
    - For detection probabilities in [0,1], prefer `loss_type="bce_logits"` and do NOT set
      `train_in_log=True` (BCE expects probability targets, not log-values).
    - `seed` fixes the validation split for a less jittery val-loss.
    """
    if checkpoint_path is None:
        raise ValueError("No checkpoint path provided, model will not be saved.")
    if not checkpoint_path.endswith(".hdf5"):
        raise ValueError("Checkpoint path must end with .hdf5")

    # --------------------------
    # Read & prepare data
    # --------------------------
    df = read_data(data_path, keys=input_keys + output_keys)

    data_X = jax.device_put(df[input_keys].to_numpy())
    data_Y = jax.device_put(df[output_keys].to_numpy())

    if train_in_log:
        # Only for regression on log(y). Avoid for BCE loss.
        data_Y = jnp.log(data_Y)
        data_Y = jnp.where(
            jnp.isneginf(data_Y),
            -jnp.finfo(jnp.result_type(float)).tiny,
            data_Y,
        )

    train_X, test_X, train_Y, test_Y = _train_test_data_split(
        data_X, data_Y, batch_size, test_size=validation_split, seed=seed
    )

    logger.info("Input Keys: " + ", ".join(input_keys))
    logger.info("Output Keys: " + ", ".join(output_keys))
    logger.info(f"Width Size: {width_size}")
    logger.info(f"Depth: {depth}")
    logger.info(f"Data Path: {data_path}")
    logger.info(f"Checkpoint Path: {checkpoint_path}")
    logger.info(f"Train Size: {len(train_X)}")
    logger.info(f"Test Size: {len(test_X)}")
    logger.info(f"Validation Split: {validation_split}")
    logger.info(f"Batch Size: {batch_size}")
    logger.info(f"Epochs: {epochs}")
    logger.info(f"Learning Rate (peak): {learning_rate}")
    logger.info(f"Loss Type: {loss_type}")
    logger.info(f"Scheduler: {'cosine+warmup' if use_cosine_decay else 'constant'}")
    logger.info(
        f"Weight Decay: {weight_decay}, Grad Clip: {grad_clip_norm}, Seed: {seed}"
    )
    if train_in_log and loss_type.lower() == "bce_logits":
        logger.warning(
            "train_in_log=True with BCE is not recommended; BCE expects probabilities."
        )

    # --------------------------
    # Model & Optimizer
    # --------------------------
    model = make_model(
        key=jrd.PRNGKey(np.random.randint(0, 2**32 - 1)),
        input_layer=len(input_keys),
        output_layer=len(output_keys),
        width_size=width_size,
        depth=depth,
    )
    model: Callable = eqx.filter_checkpoint(model)  # type: ignore

    steps_per_epoch = max(1, len(train_X) // batch_size)
    total_steps = max(1, steps_per_epoch * epochs)
    warmup_steps = min(total_steps, max(0, warmup_epochs) * steps_per_epoch)

    if use_cosine_decay:
        schedule = optax.warmup_cosine_decay_schedule(
            init_value=0.0,
            peak_value=learning_rate,
            warmup_steps=warmup_steps,
            decay_steps=max(1, total_steps - warmup_steps),
            end_value=min_lr,
        )
    else:
        schedule = learning_rate

    optimizer = optax.chain(
        optax.clip_by_global_norm(grad_clip_norm),
        optax.adamw(learning_rate=schedule, weight_decay=weight_decay),
    )

    opt_state = optimizer.init(eqx.filter(model, eqx.is_inexact_array))

    # Choose loss function
    loss_fn: Callable[..., PyTree]
    if loss_type.lower() == "mse":
        loss_fn = mse_loss_fn
    elif loss_type.lower() in ("bce", "bce_logits", "bcelogits"):
        loss_fn = bce_logits_loss_fn
    else:
        raise ValueError("loss_type must be 'mse' or 'bce_logits'")

    @eqx.filter_jit
    def make_step(
        model: eqx.nn.MLP, x: Array, y: Array, opt_state: optax.OptState
    ) -> tuple[eqx.nn.MLP, optax.OptState, Array]:
        loss, grads = loss_fn(model, x, y)
        updates, opt_state = optimizer.update(
            grads, opt_state, params=eqx.filter(model, eqx.is_inexact_array)
        )
        model = eqx.apply_updates(model, updates)
        return model, opt_state, loss

    # --------------------------
    # Training Loop
    # --------------------------
    loss_vals: list[float] = []
    val_loss_vals: list[float] = []

    with tqdm.tqdm(range(epochs), unit="epochs") as pbar:
        for epoch in pbar:
            epoch_loss = jnp.zeros(())
            # mini-batch iterate
            for i in range(0, len(train_X), batch_size):
                x = train_X[i : i + batch_size]
                y = train_Y[i : i + batch_size]
                model, opt_state, loss = make_step(model, x, y, opt_state)
                epoch_loss = epoch_loss + loss
                pbar.set_postfix({"epoch": epoch + 1, "loss": f"{loss:.5E}"})

            # epoch metrics
            loss_epoch = epoch_loss / max(1, (len(train_X) // batch_size))
            loss_vals.append(float(loss_epoch))

            val_loss, _ = loss_fn(model, test_X, test_Y)
            val_loss_vals.append(float(val_loss))

            pbar.set_description(
                f"Epoch {epoch + 1}/{epochs}, Loss: {loss_epoch:.5E}, Val Loss: {val_loss:.5E}"
            )
            if epoch % 10 == 0 or epoch == epochs - 1:
                logger.info(
                    f"Epoch {epoch + 1}/{epochs}, Loss: {loss_epoch:.5E}, Val Loss: {val_loss:.5E}"
                )

    # --------------------------
    # Save & Diagnostics
    # --------------------------
    save_model(
        filepath=checkpoint_path,
        datafilepath=data_path,
        model=model,
        names=input_keys,
        is_log=train_in_log,
    )  # type: ignore

    # Loss curves

    # Create subplots with shared x and y axes

    plt.rcParams.update({"text.usetex": True, "font.size": 18})
    _, axes = plt.subplots(3, 1, figsize=(15, 15), dpi=300, sharex=True, sharey=True)

    import glasbey

    colors = glasbey.create_palette(palette_size=2)

    # Plot 1: Training loss
    axes[0].plot(loss_vals, label="loss", color=colors[0])
    axes[0].set_yscale("log")
    axes[0].legend()
    axes[0].set_ylabel("Loss")

    # Plot 2: Validation loss
    axes[1].plot(val_loss_vals, label="val loss", color=colors[1])
    axes[1].set_yscale("log")
    axes[1].legend()
    axes[1].set_ylabel("Validation Loss")

    # Plot 3: Combined
    axes[2].plot(loss_vals, label="loss", alpha=0.7, color=colors[0])
    axes[2].plot(val_loss_vals, label="val loss", alpha=0.7, color=colors[1])
    axes[2].set_yscale("log")
    axes[2].set_xlabel("Epoch")
    axes[2].set_ylabel("Average loss per epoch")
    axes[2].legend()

    # Adjust layout and save
    plt.tight_layout()
    plt.savefig(checkpoint_path + "_loss.png")
    plt.close("all")

    # Per-instance loss distribution
    Y_hat = predict(model, data_X)
    per_instance = jnp.square(data_Y - Y_hat).squeeze(axis=-1)
    per_instance_np = np.asarray(per_instance)
    q05, q50, q95 = np.quantile(per_instance_np, [0.05, 0.5, 0.95])
    ordered = np.sort(per_instance_np)

    plt.rcParams.update({"text.usetex": True, "font.size": 16})
    plt.figure(figsize=(10, 6), dpi=300)
    colors = glasbey.create_palette(palette_size=4)
    plt.plot(ordered, label="loss per data instance", color=colors[0])
    plt.axhline(q05, linestyle="--", label="5 percent quantile", color=colors[1])
    plt.axhline(q50, linestyle="--", label="50 percent quantile", color=colors[2])
    plt.axhline(q95, linestyle="--", label="95 percent quantile", color=colors[3])
    plt.yscale("log")
    plt.xlabel(r"$x_i$")
    plt.ylabel(r"$\left(y(x_i) - \hat{y}(x_i)\right)^2$")
    plt.legend()
    plt.tight_layout()
    plt.savefig(checkpoint_path + "_total_loss.png")
    plt.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import Dict, List, Union

from gwkokab.models import BrokenPowerlawTwoPeakFull
from gwkokab.parameters import Parameters as P
from kokab.core.monk import Monk, monk_arg_parser
from kokab.utils.logger import log_info


class BrokenPowerlawTwoPeakFullMonk(Monk):
    def __init__(
        self,
        has_spin: bool,
        has_tilt: bool,
        has_eccentricity: bool,
        has_redshift: bool,
        data_filename: str,
        seed: int,
        prior_filename: str,
        poisson_mean_filename: str,
        sampler_settings_filename: str,
        n_samples: int,
        minimum_mc_error: float,
        n_checkpoints: int,
        n_max_steps: int,
        debug_nans: bool = False,
        profile_memory: bool = False,
        check_leaks: bool = False,
    ) -> None:
        self.has_spin = has_spin
        self.has_tilt = has_tilt
        self.has_eccentricity = has_eccentricity
        self.has_redshift = has_redshift

        super().__init__(
            BrokenPowerlawTwoPeakFull,
            data_filename,
            seed,
            prior_filename,
            poisson_mean_filename,
            sampler_settings_filename,
            debug_nans=debug_nans,
            profile_memory=profile_memory,
            check_leaks=check_leaks,
            analysis_name="bp2pfull",
            n_samples=n_samples,
            minimum_mc_error=minimum_mc_error,
            n_checkpoints=n_checkpoints,
            n_max_steps=n_max_steps,
        )

    @property
    def constants(self) -> Dict[str, Union[int, float, bool]]:
        return {
            "use_spin": self.has_spin,
            "use_tilt": self.has_tilt,
            "use_eccentricity": self.has_eccentricity,
            "use_redshift": self.has_redshift,
        }

    @property
    def parameters(self) -> List[str]:
        names = [P.PRIMARY_MASS_SOURCE.value, P.SECONDARY_MASS_SOURCE.value]
        if self.has_spin:
            names.append(P.PRIMARY_SPIN_MAGNITUDE.value)
            names.append(P.SECONDARY_SPIN_MAGNITUDE.value)
        if self.has_tilt:
            names.extend([P.COS_TILT_1.value, P.COS_TILT_2.value])
        if self.has_eccentricity:
            names.append(P.ECCENTRICITY.value)
        if self.has_redshift:
            names.append(P.REDSHIFT.value)
        return names

    @property
    def model_parameters(self) -> List[str]:
        all_params: List[str] = [
            "alpha1",
            "alpha2",
            "beta",
            "delta_m1",
            "delta_m2",
            "lambda_0",
            "lambda_1",
            "loc1",
            "loc2",
            "log_rate",
            "m1min",
            "m2min",
            "mbreak",
            "mmax",
            "scale1",
            "scale2",
        ]

        if self.has_spin:
            all_params.extend(["chi_loc", "chi_scale"])

        if self.has_tilt:
            all_params.extend(["cos_tilt_zeta", "cos_tilt_loc", "cos_tilt_scale"])

        if self.has_eccentricity:
            all_params.append("eccentricity_scale")

        if self.has_redshift:
            all_params.extend(["z_max", "kappa"])

        return all_params


def model_arg_parser(parser: ArgumentParser) -> ArgumentParser:
    model_group = parser.add_argument_group("Model Options")
    model_group.add_argument(
        "--add-spin",
        action="store_true",
        help="Include beta spin parameters in the model.",
    )
    model_group.add_argument(
        "--add-tilt",
        action="store_true",
        help="Include tilt parameters in the model.",
    )
    model_group.add_argument(
        "--add-eccentricity",
        action="store_true",
        help="Include eccentricity parameter in the model",
    )
    model_group.add_argument(
        "--add-redshift",
        action="store_true",
        help="Include redshift parameter in the model",
    )
    return parser


def main() -> None:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = model_arg_parser(parser)
    parser = monk_arg_parser(parser)

    args = parser.parse_args()

    log_info(start=True)

    BrokenPowerlawTwoPeakFullMonk(
        has_spin=args.add_spin,
        has_tilt=args.add_tilt,
        has_eccentricity=args.add_eccentricity,
        has_redshift=args.add_redshift,
        data_filename=args.data_filename,
        seed=args.seed,
        prior_filename=args.prior_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        n_samples=args.n_samples,
        minimum_mc_error=args.minimum_mc_error,
        n_checkpoints=args.n_checkpoints,
        n_max_steps=args.n_max_steps,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
    ).run()



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Sequence
from typing import Dict, List, Literal, Optional, Tuple, Union

import jax
from jax import lax, numpy as jnp, random as jrd
from jaxtyping import Array, PRNGKeyArray
from numpyro._typing import ConstraintT, DistributionT
from numpyro.distributions import constraints, Distribution
from numpyro.distributions.util import validate_sample
from numpyro.util import is_prng_key

from ...utils.tools import error_if, warn_if


class _LazyConstraint(constraints.Constraint):
    """A constraint that checks if a value is in the support of a
    :class:`LazyJointDistribution`.
    """

    def __init__(
        self,
        *marginal_distributions: Union[DistributionT, jax.tree_util.Partial],
        dependencies: Dict[int, Dict[str, int]],
        event_slices: Sequence[int | Tuple[int, int]],
    ) -> None:
        self.marginal_distributions = tuple(marginal_distributions)
        self.dependencies = dependencies
        self.event_slices = event_slices

    def __call__(self, x: Array) -> Array:
        marginal_dists = list(self.marginal_distributions)
        for i, dep in self.dependencies.items():
            kwargs = {
                k: jax.lax.dynamic_index_in_dim(x, v, axis=-1, keepdims=False)
                for k, v in dep.items()
            }
            mdist = marginal_dists[i]
            marginal_dists[i] = mdist.func(*mdist.args, **mdist.keywords, **kwargs)  # type: ignore
        mask = None
        for mdist, event_slice in zip(marginal_dists, self.event_slices, strict=True):
            constraint: ConstraintT = mdist.support  # type: ignore
            if isinstance(event_slice, int):
                x_slice = lax.dynamic_index_in_dim(
                    x, event_slice, axis=-1, keepdims=False
                )
            else:
                x_slice = lax.dynamic_slice_in_dim(
                    x,
                    event_slice[0],
                    event_slice[1] - event_slice[0],
                    axis=-1,
                )
            if mask is None:
                mask = constraint.check(x_slice)
            else:
                mask = jnp.logical_and(mask, constraint.check(x_slice))
        return mask  # type: ignore

    def tree_flatten(self):
        return (self.marginal_distributions, self.dependencies, self.event_slices), (
            ("marginal_distributions", "dependencies", "event_slices"),
            dict(),
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _LazyConstraint):
            return False
        return (
            all(
                constraint == other_constraint
                for constraint, other_constraint in zip(
                    self.marginal_distributions,
                    other.marginal_distributions,
                    strict=True,
                )
            )
            and self.dependencies == other.dependencies
            and self.event_slices == other.event_slices
        )


class LazyJointDistribution(Distribution):
    pytree_aux_fields = ("shaped_values", "partial_order", "dependencies")
    pytree_data_fields = ("_support", "marginal_distributions")

    def __init__(
        self,
        *marginal_distributions: Union[DistributionT, jax.tree_util.Partial],
        dependencies: Dict[int, Dict[str, int]],
        partial_order: List[int],
        dependencies_event_shape: Optional[List[Tuple[int, ...]]] = None,
        flatten_method: Optional[Literal["deep", "shallow"]] = None,
        support: Optional[ConstraintT] = None,
        validate_args: Optional[bool] = None,
    ) -> None:
        """Construct a joint distribution from one or more marginal distributions.

        You may pass individual `Distribution` instances or nest them inside
        :class:`LazyJointDistribution`s. The `flatten_method` argument allows flattening of nested
        joints into a single flat list of marginals.

        Parameters
        ----------
        marginal_distributions : *Union[DistributionT, jax.tree_util.Partial]
            One or more marginal distributions. Each marginal distribution can be an instance of
            `numpyro.distributions.Distribution` or a `jax.tree_util.Partial` that returns a
            `numpyro.distributions.Distribution` when called with its arguments.
        dependencies : Dict[int, Dict[str, int]]
            A dictionary mapping the index of each marginal distribution that is a
            `jax.tree_util.Partial` to another dictionary that maps the names of its dependency
            parameters to the indices of the marginal distributions they depend on.
            This is used to specify which variables each lazy variable depends on.
        partial_order : Optional[Tuple[str, int, int]], optional
            A tuple defining a partial order for the lazy variables.
            Each entry in the tuple should be of the form (var_name, event_index, marginal_index),
            such that elements coming earlier in the tuple are not dependent on elements coming later.
            This is used to ensure that when sampling from the joint distribution, the lazy variables
            are sampled in an order that respects their dependencies.
        dependencies_event_shape : Optional[List[Tuple[int, ...]]], optional
            A list of event shapes for the dependencies of each marginal distribution.
            This is used to validate the shapes of the dependency variables when constructing
        flatten_method : Optional[Literal[&quot;deep&quot;, &quot;shallow&quot;]], optional
            Currently not used.
            If "shallow", one level of nested :class:`LazyJointDistributions` will be flattened.
            If "deep", all levels of nested :class:`LazyJointDistributions` will be recursively flattened.
            If None (default), the nesting is preserved as-is.
        support : Optional[ConstraintT], optional
            The constraint object representing the support of the joint distribution.
            If not provided, it is computed from the support of the marginals.
        validate_args : Optional[bool], optional
            Whether to validate distribution parameters and inputs. Default is None.

        Raises
        ------
        ValueError
             If no marginal distributions are provided.
        """
        error_if(
            not marginal_distributions,
            msg="At least one marginal distribution is required.",
        )
        error_if(
            len(partial_order) == 0,
            msg="`partial_order` must be provided. If there is not lazy "
            "variable then use a standard `gwkokab.models.JointDistribution` instead.",
        )
        error_if(
            dependencies is None,
            msg="`dependencies` must be provided. If there is not lazy "
            "variable then use a standard `gwkokab.models.JointDistribution` instead.",
        )
        error_if(
            len(partial_order) != len(dependencies),
            msg="`partial_order` and `dependencies` must have the same length.",
        )
        warn_if(
            flatten_method is not None,
            msg="The `flatten_method` argument is not used with in "
            "`LazyJointDistribution`. It will be ignored.",
        )

        # TODO(Qazalbash): Implement flattening logic
        # marginal_flatten = _flatten_marginal_distributions(
        #     marginal_distributions, flatten_method
        # )
        marginal_flatten = marginal_distributions
        self.marginal_distributions: Sequence[
            Union[DistributionT, jax.tree_util.Partial]
        ] = tuple(marginal_flatten)
        self.shaped_values: Sequence[Union[int, Tuple[int, int]]] = tuple()
        dependencies = dependencies or {}  # for type checker
        k = 0
        for i, d in enumerate(marginal_flatten):
            error_if(
                not isinstance(d, (Distribution, jax.tree_util.Partial)),
                msg="All marginals must be instances of "
                "`numpyro.distributions.Distribution` and `jax.tree_util.Partial`. "
                f"Got {type(d)}",
            )
            if isinstance(d, jax.tree_util.Partial):
                event_shape = (
                    dependencies_event_shape[i] if dependencies_event_shape else ()
                )
            else:
                event_shape = d.event_shape
            if event_shape:
                self.shaped_values += ((k, k + event_shape[0]),)
                k += event_shape[0]
            else:
                self.shaped_values += (k,)
                k += 1
        batch_shape = ()
        if support is None:
            self._support = _LazyConstraint(
                *self.marginal_distributions,
                dependencies=dependencies,
                event_slices=self.shaped_values,
            )
        else:
            self._support = support

        self.partial_order = partial_order
        self.dependencies = dependencies

        super(LazyJointDistribution, self).__init__(
            batch_shape=batch_shape,
            event_shape=(k,),
            validate_args=validate_args,
        )

    @constraints.dependent_property(is_discrete=False)
    def support(self) -> constraints.Constraint:
        """The support of the joint distribution."""
        return self._support

    @validate_sample
    def log_prob(self, value: Array) -> Array:
        marginal_dists: List[DistributionT] = list(self.marginal_distributions)  # type: ignore
        for i, dep in self.dependencies.items():
            kwargs = {
                k: jax.lax.dynamic_index_in_dim(value, v, axis=-1, keepdims=False)
                for k, v in dep.items()
            }
            mdist: jax.tree_util.Partial = marginal_dists[i]
            marginal_dists[i] = mdist.func(*mdist.args, **mdist.keywords, **kwargs)  # type: ignore
        log_prob_val = jnp.zeros(value.shape[:-1], dtype=value.dtype)
        for m_dist, event_slice in zip(marginal_dists, self.shaped_values, strict=True):
            if isinstance(event_slice, int):
                value_slice = lax.dynamic_index_in_dim(
                    value, event_slice, axis=-1, keepdims=False
                )
            else:
                value_slice = lax.dynamic_slice_in_dim(
                    value,
                    event_slice[0],
                    event_slice[1] - event_slice[0],
                    axis=-1,
                )
            log_prob_val += m_dist.log_prob(value_slice)
        return log_prob_val

    def sample(self, key: PRNGKeyArray, sample_shape: tuple[int, ...] = ()):
        assert is_prng_key(key)
        n_total = len(self.marginal_distributions)
        independent = set(range(n_total)) - set(self.partial_order)

        samples = [jnp.empty((*sample_shape, 1)) for _ in range(n_total)]

        for i in independent:
            mdist: DistributionT = self.marginal_distributions[i]  # type: ignore
            key, subkey = jrd.split(key)
            samples[i] = mdist.sample(subkey, sample_shape).reshape(*sample_shape, -1)

        for i in self.partial_order:
            key, subkey = jrd.split(key)
            kwargs = {k: samples[v] for k, v in self.dependencies[i].items()}
            dist = self.marginal_distributions[i]
            if isinstance(dist, jax.tree_util.Partial):
                dist = dist.func(*dist.args, **dist.keywords, **kwargs)  # type: ignore
            samples[i] = dist.sample(
                subkey, sample_shape[dist.event_dim + 1 :]
            ).reshape(*sample_shape, -1)

        samples_concatenated = jnp.concatenate(samples, axis=-1)
        return samples_concatenated



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple

import jax
from jax import numpy as jnp
from jaxtyping import Array
from numpyro._typing import DistributionT


def variance_of_single_event_likelihood(
    model_instance: DistributionT,
    n_buckets: int,
    data_group: Tuple[Array, ...],
    log_ref_priors_group: Tuple[Array, ...],
    masks_group: Tuple[Array, ...],
) -> Array:
    """See equation 9 and 10 in https://arxiv.org/abs/2406.16813."""

    @jax.jit
    def _variance_of_single_event_likelihood(*args: Array):
        data_group = args[0:n_buckets]
        log_ref_priors_group = args[n_buckets : 2 * n_buckets]
        masks_group = args[2 * n_buckets : 3 * n_buckets]

        variance = jnp.zeros(())
        # Σ log Σ exp (log p(θ|data_n) - log π_n)
        for batched_data, batched_log_ref_priors, batched_masks in zip(
            data_group, log_ref_priors_group, masks_group
        ):
            safe_data = jnp.where(
                jnp.expand_dims(batched_masks, axis=-1),
                batched_data,
                model_instance.support.feasible_like(batched_data),
            )
            safe_log_ref_prior = jnp.where(batched_masks, batched_log_ref_priors, 0.0)

            n_events_per_bucket, n_samples, _ = batched_data.shape
            batched_model_log_prob = jax.vmap(
                jax.vmap(model_instance.log_prob, axis_size=n_samples),
                axis_size=n_events_per_bucket,
            )(safe_data)  # type: ignore
            safe_model_log_prob = jnp.where(
                batched_masks, batched_model_log_prob, -jnp.inf
            )
            batched_log_prob: Array = safe_model_log_prob - safe_log_ref_prior
            safe_batched_log_prob = jnp.where(
                batched_masks & (~jnp.isnan(batched_log_prob)),
                batched_log_prob,
                -jnp.inf,
            )
            log_prob_sum = jax.nn.logsumexp(
                safe_batched_log_prob,
                axis=-1,
                where=~jnp.isneginf(safe_batched_log_prob),
            )
            log_prob_sum_2 = jax.nn.logsumexp(
                2.0 * safe_batched_log_prob,
                axis=-1,
                where=~jnp.isneginf(safe_batched_log_prob),
            )
            safe_prob_sum = jnp.where(
                jnp.isneginf(log_prob_sum), 0.0, jnp.exp(log_prob_sum)
            )
            safe_prob_sum_2 = jnp.where(
                jnp.isneginf(log_prob_sum_2), 0.0, jnp.exp(log_prob_sum_2)
            )

            N_pe = jnp.count_nonzero(batched_masks, axis=-1)

            variance += (safe_prob_sum_2 / safe_prob_sum**2 - 1.0 / N_pe).sum()

        return variance

    return _variance_of_single_event_likelihood(
        *data_group, *log_ref_priors_group, *masks_group
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


#
"""Literals used in Kokab."""

from typing import Final, Literal


LOG_REF_PRIOR_NAME: Final[Literal["log_prior"]] = "log_prior"
"""Helper variable to store the column name of log reference prior."""


INFERENCE_DIRECTORY: Final[Literal["inference"]] = "inference"
"""Name of the directory to store inference results."""


POSTERIOR_SAMPLES_FILENAME: Final[Literal["samples.dat"]] = "samples.dat"
"""Name of the file to store posterior samples."""



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Callable, Sequence
from typing import Optional, Tuple, Union

import equinox as eqx
import jax
from jax import numpy as jnp
from jaxtyping import Array, PRNGKeyArray

from ..models.utils import ScaledMixture
from ..utils.tools import error_if
from ..utils.train import load_model


def poisson_mean_from_neural_vt(
    key: PRNGKeyArray,
    parameters: Sequence[str],
    filename: str,
    batch_size: Optional[int] = None,
    num_samples: int = 1_000,
    time_scale: Union[int, float, Array] = 1.0,
) -> Tuple[
    Optional[Callable[[Array], Array]],
    Callable[[ScaledMixture], Array],
    float | Array,
    Callable[[ScaledMixture], Array],
]:
    error_if(not parameters, msg="parameters sequence cannot be empty")
    error_if(
        not isinstance(parameters, Sequence),
        TypeError,
        f"parameters must be a Sequence, got {type(parameters)}",
    )
    error_if(
        not all(isinstance(p, str) for p in parameters),
        TypeError,
        "all parameters must be strings",
    )
    if batch_size is not None:
        error_if(
            not isinstance(batch_size, int),
            TypeError,
            f"batch_size must be an integer, got {type(batch_size)}",
        )
        error_if(
            batch_size < 1,
            msg=f"batch_size must be a positive integer, got {batch_size}",
        )

    names, neural_vt_model = load_model(filename)
    error_if(
        any(name not in parameters for name in names),
        msg=f"Model in {filename} expects parameters {names}, but received "
        f"{parameters}. Missing: {set(names) - set(parameters)}",
    )

    shuffle_indices = [parameters.index(name) for name in names]

    @jax.jit
    def log_vt(x: Array) -> Array:
        x_new = x[..., shuffle_indices]
        return jnp.squeeze(
            jax.lax.map(neural_vt_model, x_new, batch_size=batch_size), axis=-1
        )

    def _poisson_mean(scaled_mixture: ScaledMixture) -> Array:
        component_sample = scaled_mixture.component_sample(key, (num_samples,))
        # vmapping over components
        log_vt_values = jax.vmap(log_vt, in_axes=1)(component_sample)
        mean_per_component = jnp.exp(
            scaled_mixture.log_scales + jax.nn.logsumexp(log_vt_values, axis=-1)
        )
        return (time_scale / num_samples) * jnp.sum(mean_per_component, axis=-1)

    @eqx.filter_jit
    def _variance_of_estimator(scaled_mixture: ScaledMixture) -> Array:
        component_sample = scaled_mixture.component_sample(key, (num_samples,))
        # vmapping over components
        log_vt_values = jax.vmap(log_vt, in_axes=1)(component_sample)
        term2 = jnp.exp(
            2.0 * jnp.log(time_scale)
            - 3.0 * jnp.log(num_samples)
            + 2.0 * scaled_mixture.log_scales
            + 2.0 * jax.nn.logsumexp(log_vt_values, axis=-1)
        )
        term1 = jnp.exp(
            2.0 * jnp.log(time_scale)
            - 2.0 * jnp.log(num_samples)
            + 2.0 * scaled_mixture.log_scales
            + jax.nn.logsumexp(2.0 * log_vt_values, axis=-1)
        )
        return jnp.sum(term1 - term2, axis=-1)

    return log_vt, _poisson_mean, time_scale, _variance_of_estimator
